from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IWCContainerLocator
from plone.app.iterate.testing import PLONEAPPITERATEDEX_INTEGRATION_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from zope.annotation.interfaces import IAnnotatable
from zope.annotation.interfaces import IAnnotations
from zope.component import getAdapters

import unittest


class AnnotationsTestCase(unittest.TestCase):
    layer = PLONEAPPITERATEDEX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Document", "s1")
        self.s1 = self.portal["s1"]

    def test_object_annotatable(self):
        self.assertTrue(IAnnotatable.providedBy(self.s1))

    def test_annotation_saved_on_checkin(self):
        # First we get and save a custom annotation to the existing object
        obj_annotations = IAnnotations(self.s1)
        self.assertEqual(obj_annotations, {})

        obj_annotations["key1"] = "value1"
        obj_annotations = IAnnotations(self.s1)
        self.assertEqual(obj_annotations, {"key1": "value1"})

        # Now, let's get a working copy for it.
        locators = getAdapters((self.s1,), IWCContainerLocator)
        location = "plone.app.iterate.parent"
        locator = [c[1] for c in locators if c[0] == location][0]

        policy = ICheckinCheckoutPolicy(self.s1)

        wc = policy.checkout(locator())

        # Annotations should be the same
        new_annotations = IAnnotations(wc)
        self.assertEqual(new_annotations["key1"], "value1")

        # Now, let's modify the existing one, and create a new one
        new_annotations["key1"] = "value2"
        new_annotations["key2"] = "value1"

        # Check that annotations were stored correctly and original ones were
        # not overriten
        new_annotations = IAnnotations(wc)
        self.assertEqual(new_annotations["key1"], "value2")
        self.assertEqual(new_annotations["key2"], "value1")

        obj_annotations = IAnnotations(self.s1)
        self.assertEqual(obj_annotations["key1"], "value1")
        self.assertFalse("key2" in obj_annotations)

        # Now, we do a checkin
        policy = ICheckinCheckoutPolicy(wc)
        policy.checkin("Commit message")

        # And finally check that the old object has the same annotations as
        # its working copy

        obj_annotations = IAnnotations(self.s1)
        self.assertTrue("key1" in obj_annotations)
        self.assertTrue("key2" in obj_annotations)
        self.assertEqual(obj_annotations["key1"], "value2")
        self.assertEqual(obj_annotations["key2"], "value1")



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from .interfaces import annotation_key
from persistent.mapping import PersistentMapping
from zope.annotation import IAnnotations


def get_storage(context, default=None):
    annotations = IAnnotations(context)
    if annotation_key not in annotations:
        if default is not None:
            return default
        annotations[annotation_key] = PersistentMapping()
    return annotations[annotation_key]



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from Acquisition import aq_inner
from plone.app.iterate import PloneMessageFactory as _
from plone.app.iterate.interfaces import CheckinException
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from zope.component import getMultiAdapter


class Checkin(BrowserView):
    index = ViewPageTemplateFile("checkin.pt")

    def __call__(self):
        context = aq_inner(self.context)

        if "form.button.Checkin" in self.request.form:
            control = getMultiAdapter((context, self.request), name="iterate_control")
            if not control.checkin_allowed():
                raise CheckinException("Not a checkout")

            message = self.request.form.get("checkin_message", "")

            policy = ICheckinCheckoutPolicy(context)
            baseline = policy.checkin(message)

            IStatusMessage(self.request).addStatusMessage(_("Checked in"), type="info")
            view_url = baseline.restrictedTraverse("@@plone_context_state").view_url()
            self.request.response.redirect(view_url)
        elif "form.button.Cancel" in self.request.form:
            view_url = context.restrictedTraverse("@@plone_context_state").view_url()
            self.request.response.redirect(view_url)
        else:
            return self.index()



from plone.app.contenttypes.interfaces import IDocument
from plone.dexterity.content import Container
from zope.interface import implementer
from zope.interface import Interface


class IFolderishDocument(Interface):
    """Marker interface"""


@implementer(IDocument, IFolderishDocument)
class FolderishDocument(Container):
    """A Dexterity based test type containing a set of standard fields."""

    pass



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
"""
from . import interfaces
from zope.event import notify
from zope.interface import implementer
from zope.interface.interfaces import ObjectEvent


@implementer(interfaces.ICheckoutEvent)
class CheckoutEvent(ObjectEvent):
    def __init__(self, baseline, wc, relation):
        ObjectEvent.__init__(self, baseline)
        self.working_copy = wc
        self.relation = relation


@implementer(interfaces.ICheckinEvent)
class CheckinEvent(ObjectEvent):
    def __init__(self, wc, baseline, relation, message):
        ObjectEvent.__init__(self, wc)
        self.baseline = baseline
        self.relation = relation
        self.message = message


@implementer(interfaces.IAfterCheckinEvent)
class AfterCheckinEvent(ObjectEvent):
    def __init__(self, new_baseline, checkin_message):
        super().__init__(new_baseline)
        self.message = checkin_message


@implementer(interfaces.ICancelCheckoutEvent)
class CancelCheckoutEvent(ObjectEvent):
    def __init__(self, wc, baseline):
        ObjectEvent.__init__(self, wc)
        self.baseline = baseline


@implementer(interfaces.IWorkingCopyDeletedEvent)
class WorkingCopyDeletedEvent(ObjectEvent):
    def __init__(self, wc, baseline, relation):
        ObjectEvent.__init__(self, wc)
        self.baseline = baseline
        self.relation = relation


@implementer(interfaces.IBeforeCheckoutEvent)
class BeforeCheckoutEvent(ObjectEvent):
    pass


def handleDeletion(reference, event):
    # a filtering/enriching event rebroadcaster for working copy deletions
    workingCopy = reference.getSourceObject()
    baseline = reference.getTargetObject()
    notify(WorkingCopyDeletedEvent(workingCopy, baseline, reference))



from plone.app.iterate.testing import PLONEAPPITERATEDEX_FUNCTIONAL_TESTING
from plone.testing import layered
from unittest import TestSuite

import doctest


def test_suite():
    suite = TestSuite()
    OPTIONFLAGS = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    suite.addTest(
        layered(
            doctest.DocFileSuite(
                "dexterity.rst",
                optionflags=OPTIONFLAGS,
                package="plone.app.iterate.tests",
            ),
            layer=PLONEAPPITERATEDEX_FUNCTIONAL_TESTING,
        )
    )
    return suite



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
from plone.app.iterate import interfaces


def handleBeforeCheckout(event):
    archiver = interfaces.IObjectArchiver(event.object)
    if not archiver.isVersionable():
        return
    if archiver.isModified() or not archiver.isVersioned():
        archiver.save("Baseline created")


def handleAfterCheckin(event):
    archiver = interfaces.IObjectArchiver(event.object)
    if not archiver.isVersionable():
        return
    archiver.save(event.message)



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from Acquisition import aq_inner
from plone.app.iterate import PloneMessageFactory as _
from plone.app.iterate.interfaces import CheckoutException
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IWCContainerLocator
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from zope.component import getAdapters
from zope.component import getMultiAdapter


class Checkout(BrowserView):
    index = ViewPageTemplateFile("checkout.pt")

    def containers(self):
        """Get a list of potential containers"""
        context = aq_inner(self.context)
        for name, locator in getAdapters((context,), IWCContainerLocator):
            if locator.available:
                yield dict(name=name, locator=locator)

    def __call__(self):
        context = aq_inner(self.context)

        # We want to redirect to a specific template, else we might
        # end up downloading a file
        if "form.button.Checkout" in self.request.form:
            control = getMultiAdapter((context, self.request), name="iterate_control")
            if not control.checkout_allowed():
                raise CheckoutException("Not allowed")

            location = self.request.form.get("checkout_location", None)
            locator = None
            try:
                locator = [
                    c["locator"] for c in self.containers() if c["name"] == location
                ][0]
            except IndexError:
                IStatusMessage(self.request).addStatusMessage(
                    _("Cannot find checkout location"), type="error"
                )
                view_url = context.restrictedTraverse(
                    "@@plone_context_state"
                ).view_url()
                self.request.response.redirect(view_url)
                return

            policy = ICheckinCheckoutPolicy(context)
            wc = policy.checkout(locator())

            # we do this for metadata update side affects which will update
            # lock info
            context.reindexObject(idxs=["review_state"])

            IStatusMessage(self.request).addStatusMessage(
                _("Check-out created"), type="info"
            )
            view_url = wc.restrictedTraverse("@@plone_context_state").view_url()
            self.request.response.redirect(view_url)
        elif "form.button.Cancel" in self.request.form:
            view_url = context.restrictedTraverse("@@plone_context_state").view_url()
            self.request.response.redirect(view_url)
        else:
            return self.index()



##################################################################
#
# (C) Copyright 2006-2007 ObjectRealms, LLC
# (C) Copyright 2007-2017 Plone Foundation
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with CMFDeployment; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################
"""
Base Checkin Checkout Policy For Content

"""

from Acquisition import aq_inner
from Acquisition import aq_parent
from plone.app.iterate import interfaces
from plone.app.iterate.event import BeforeCheckoutEvent
from plone.app.iterate.event import CancelCheckoutEvent
from plone.app.iterate.event import CheckoutEvent
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IObjectCopier
from plone.app.iterate.util import get_storage
from zc.relation.interfaces import ICatalog
from zope import component
from zope.component import queryAdapter
from zope.event import notify
from zope.interface import implementer


@implementer(ICheckinCheckoutPolicy)
class CheckinCheckoutBasePolicyAdapter:
    """Base Checkin Checkout Policy For Content

    - on checkout context is the baseline
    - on checkin context is the working copy.

    """

    # used when creating baseline version for first time
    default_base_message = "Created Baseline"

    def __init__(self, context):
        self.context = context

    def checkin(self, checkin_message):
        raise NotImplementedError()

    def checkout(self, container):
        # see interface
        notify(BeforeCheckoutEvent(self.context))

        # use the object copier to checkout the content to the container
        copier = queryAdapter(self.context, IObjectCopier)
        # The container for the Portal's working copy is the Portal itself.
        if self.context.portal_type == "Plone Site":
            working_copy_container = self.context
        else:
            working_copy_container = container
        working_copy, relation = copier.copyTo(working_copy_container)

        # publish the event for any subscribers
        notify(CheckoutEvent(self.context, working_copy, relation))

        # finally return the working copy
        return working_copy

    def cancelCheckout(self):
        # see interface

        # get the baseline
        baseline = self._getBaseline()

        # publish an event
        notify(CancelCheckoutEvent(self.context, baseline))

        # Remove the relationship
        relation = self._get_relation_to_baseline()
        catalog = component.queryUtility(ICatalog)
        catalog.unindex(relation)

        # delete the working copy
        wc_container = aq_parent(aq_inner(self.context))
        wc_container.manage_delObjects([self.context.getId()])

        return baseline

    #################################
    #  Checkin Support Methods

    def getBaseline(self):
        raise NotImplementedError()

    def getWorkingCopy(self):
        raise NotImplementedError()

    def getProperties(self, obj, default=None):
        return get_storage(obj, default=default)


@implementer(interfaces.IObjectCopier)
@component.adapter(interfaces.IIterateAware)
class BaseContentCopier:
    def __init__(self, context):
        self.context = context

    def _copyBaseline(self, container):
        # copy the context from source to the target container
        source_container = aq_parent(aq_inner(self.context))
        clipboard = source_container.manage_copyObjects([self.context.getId()])
        result = container.manage_pasteObjects(clipboard)

        # get a reference to the working copy
        target_id = result[0]["new_id"]
        target = container._getOb(target_id)
        return target



from Acquisition import aq_base
from Acquisition import aq_inner
from plone.app.iterate.dexterity import ITERATE_RELATION_NAME
from zc.relation.interfaces import ICatalog
from zope import component
from zope.intid.interfaces import IIntIds


def get_relations(context):
    context = aq_inner(context)
    # get id
    intids = component.getUtility(IIntIds)
    id = intids.queryId(aq_base(context))
    if not id:
        # for objects without intid or
        # objects being deleted in the current transaction return empty list
        return []
    # ask catalog
    catalog = component.getUtility(ICatalog)
    relations = list(catalog.findRelations({"to_id": id}))
    relations += list(catalog.findRelations({"from_id": id}))
    relations = filter(lambda r: r.from_attribute == ITERATE_RELATION_NAME, relations)
    return list(relations)


def get_checkout_relation(context):
    relations = get_relations(context)
    if len(relations) > 0:
        return relations[0]
    else:
        return None


def get_baseline(context):
    relation = get_checkout_relation(context)
    if relation and relation.from_id:
        intids = component.getUtility(IIntIds)
        context_id = intids.queryId(aq_base(context))
        if context_id == relation.from_id:
            return None
        return intids.getObject(relation.from_id)
    return None


def get_working_copy(context):
    relation = get_checkout_relation(context)
    if relation and relation.to_id:
        intids = component.getUtility(IIntIds)
        return intids.getObject(relation.to_id)
    return None



from plone.base.interfaces.installable import INonInstallable
from zope.interface import implementer


@implementer(INonInstallable)
class HiddenProfiles:
    def getNonInstallableProfiles(self):
        """Prevents uninstall profile from showing up in the profile list
        when creating a Plone site.

        """
        return [
            "plone.app.iterate:uninstall",
            "plone.app.iterate:to1000",
        ]



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "6.1.0"

long_description = (
    f"{Path('README.rst').read_text()}\n{Path('CHANGES.rst').read_text()}"
)

setup(
    name="plone.app.iterate",
    version=version,
    description="check-out/check-in staging for Plone",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    # Get more strings from
    # https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="check-out check-in staging",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://pypi.org/project/plone.app.iterate",
    license="GPL version 2",
    packages=find_packages(),
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "Acquisition",
        "DateTime",
        "plone.app.relationfield",
        "plone.locking",
        "plone.memoize",
        "Products.CMFEditions",
        "Products.CMFPlacefulWorkflow",
        "Products.DCWorkflow",
        "Products.GenericSetup>=1.8.2",
        "Products.statusmessages",
        "setuptools",
        "Zope",
        "plone.app.layout",
        "plone.base",
        "plone.dexterity",
        "plone.registry",
        "z3c.relationfield",
        "zc.relation",
        "zope.intid",
    ],
    extras_require={
        "test": [
            "plone.app.testing",
            "plone.app.contenttypes[test]",
            "plone.testing",
        ],
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



##################################################################
#
# (C) Copyright 2006 ObjectRealms, LLC
# All Rights Reserved
#
# This file is part of iterate.
#
# iterate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# iterate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with iterate; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
##################################################################

from AccessControl import getSecurityManager
from plone.app.iterate.browser.control import Control
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.testing import PLONEAPPITERATEDEX_INTEGRATION_TESTING
from plone.app.testing import login
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.dexterity.utils import createContentInContainer
from plone.locking.interfaces import ILockable
from z3c.relationfield import RelationValue
from zc.relation.interfaces import ICatalog
from zope import component
from zope.annotation.interfaces import IAnnotations
from zope.intid.interfaces import IIntIds

import unittest


class TestIterations(unittest.TestCase):
    """Test plone.app.iterate on containers."""

    layer = PLONEAPPITERATEDEX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        login(self.portal, TEST_USER_NAME)

        self.wf = self.portal.portal_workflow
        self.wf.setChainForPortalTypes(
            ("FolderishDocument",), "simple_publication_workflow"
        )

        # add a folder with two documents in it
        self.portal.invokeFactory("Folder", "docs")
        self.portal.docs.invokeFactory("FolderishDocument", "doc1")
        self.portal.docs.invokeFactory("FolderishDocument", "doc2")

        # add a working copy folder
        self.portal.invokeFactory("Folder", "workarea")

        self.repo = self.portal.portal_repository

    def get_control_view(self, doc):
        # Remove any memoized items from the request.
        # Especially, we may have memoized 'cancel_allowed'.
        request = self.layer["request"]
        IAnnotations(request).pop("plone.memoize", None)
        return Control(doc, request)

    def test_container_workflowState(self):
        # ensure baseline workflow state is retained on checkin, including
        # security

        doc = self.portal.docs.doc1

        # sanity check that owner can edit visible docs
        setRoles(self.portal, TEST_USER_ID, ["Owner"])
        self.assertTrue(
            getSecurityManager().checkPermission(
                "Modify portal content", self.portal.docs.doc1
            )
        )

        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.wf.doActionFor(doc, "publish")
        state = self.wf.getInfoFor(doc, "review_state")

        self.repo.save(doc)
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        wc_state = self.wf.getInfoFor(wc, "review_state")

        self.assertNotEqual(state, wc_state)

        baseline = ICheckinCheckoutPolicy(wc).checkin("modified")
        bstate = self.wf.getInfoFor(baseline, "review_state")
        self.assertEqual(bstate, "published")

        self.assertEqual(len(self.portal.workarea.objectIds()), 0)
        setRoles(self.portal, TEST_USER_ID, ["Owner"])

    def test_container_baselineCreated(self):
        # if a baseline has no version ensure that one is created on checkout
        self.assertIn("FolderishDocument", self.repo.getVersionableContentTypes())

        doc = self.portal.docs.doc1
        self.assertTrue(self.repo.isVersionable(doc))

        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)

        self.assertEqual(wc.id, "working_copy_of_doc1")
        self.assertIn("working_copy_of_doc1", self.portal.workarea.objectIds())

    def test_container_folderOrder(self):
        """When an item is checked out and then back in, the original
        folder order is preserved."""
        container = self.portal.docs
        doc = container.doc1
        original_position = container.getObjectPosition(doc.getId())

        # check that there is another document which could interact with
        # position of document the test work on
        doc2_position = container.getObjectPosition("doc2")
        self.assertTrue(doc2_position > original_position)

        self.repo.save(doc)
        wc = ICheckinCheckoutPolicy(doc).checkout(container)
        wc.text = "new document text"

        # check that the copy is put after the second document
        copy_position = container.getObjectPosition(wc.getId())
        self.assertTrue(copy_position > doc2_position)

        new_doc = ICheckinCheckoutPolicy(wc).checkin("updated")
        new_position = container.getObjectPosition(new_doc.getId())
        self.assertEqual(new_position, original_position)

    def test_container_contents(self):
        """When an folderish content is checked out, and item is added, and then
        the content is checked back in, if resources (images or files) are added item is in the new
        version of the folder they are moved to the baseline container.  UIDs of contained content are also
        preserved."""
        container = self.portal.docs
        doc = container.doc1

        wc = ICheckinCheckoutPolicy(doc).checkout(container)
        self.repo.save(wc)
        image = createContentInContainer(
            wc,
            "Image",
            id="an-image",
        )
        another_doc = createContentInContainer(
            wc,
            "FolderishDocument",
            id="document",
        )
        just_another_nested_doc = createContentInContainer(
            another_doc,
            "FolderishDocument",
            id="just_another_doc",
        )
        image_uid = image.UID()
        another_doc_uid = another_doc.UID()
        just_another_nested_doc_uid = just_another_nested_doc.UID()

        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")

        self.assertTrue("an-image" in baseline)
        self.assertTrue("document" in baseline)
        self.assertTrue("just_another_doc" in baseline["document"])
        self.assertEqual(baseline["an-image"].UID(), image_uid)
        self.assertEqual(baseline["document"].UID(), another_doc_uid)
        self.assertEqual(
            baseline["document"]["just_another_doc"].UID(), just_another_nested_doc_uid
        )

    def test_container_contents_with_collisions(self):
        """When an folderish content is checked out, and item is added, and then
        the content is checked back in, if resources (images or files) are added item is in the new
        version of the folder they are moved to the baseline container.  UIDs of contained content are also
        preserved and collisions handled by Zope cut/paste machinery."""
        container = self.portal.docs
        baseline = container.doc1
        baseline_image = createContentInContainer(
            baseline,
            "Image",
            id="an-image",
        )
        self.repo.save(baseline)
        baseline_image_uid = baseline_image.UID()

        wc = ICheckinCheckoutPolicy(baseline).checkout(container)
        self.repo.save(wc)
        image = createContentInContainer(
            wc,
            "Image",
            id="an-image",
        )
        image_uid = image.UID()

        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")

        self.assertTrue("copy_of_an-image" in baseline)
        self.assertTrue("an-image" in baseline)
        self.assertEqual(baseline["an-image"].UID(), baseline_image_uid)
        self.assertEqual(baseline["copy_of_an-image"].UID(), image_uid)

    def test_container_default_page_is_kept_in_folder(self):
        # Ensure that a default page that is checked out and back in is still
        # the default page.
        folder = self.portal.docs
        doc = folder.doc1
        from Products.CMFDynamicViewFTI.interfaces import ISelectableBrowserDefault

        ISelectableBrowserDefault(folder).setDefaultPage("doc1")
        self.assertEqual(folder.getProperty("default_page", ""), "doc1")
        self.assertEqual(folder.getDefaultPage(), "doc1")
        # Note: when checking out to self.portal.workarea it surprisingly works
        # without changes.  But the default behavior in Plone is to check a
        # document out in its original folder, so that is what we check here.
        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        doc = ICheckinCheckoutPolicy(wc).checkin("updated")
        self.assertEqual(folder.getProperty("default_page", ""), "doc1")
        self.assertEqual(folder.getDefaultPage(), "doc1")

    def test_container_control_portal(self):
        # The portal is not versionable, but it is lockable.
        doc = self.portal
        self.assertNotIn("Plone Site", self.repo.getVersionableContentTypes())
        self.assertFalse(self.repo.isVersionable(doc))
        try:
            lockable = ILockable(doc)
        except TypeError:
            lockable = False
        self.assertTrue(lockable)

        # Initially of course you can only check-out.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertTrue(control.checkout_allowed())

        # Make a checkout.  Now you can only check-in or cancel.
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        control = self.get_control_view(wc)
        self.assertTrue(control.checkin_allowed())
        self.assertTrue(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

        # On the original you cannot do anything now, because there is a working copy.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

    def test_container_control_lockable_folder(self):
        # LockableFolder is not versionable, but it is lockable.
        self.portal.invokeFactory("LockableFolder", "safe")
        doc = self.portal.safe
        self.assertNotIn("LockableFolder", self.repo.getVersionableContentTypes())
        self.assertFalse(self.repo.isVersionable(doc))
        try:
            lockable = ILockable(doc)
        except TypeError:
            lockable = False
        self.assertTrue(lockable)

        # Initially of course you can only check-out.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertTrue(control.checkout_allowed())

        # Make a checkout.  Now you can only check-in or cancel.
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        control = self.get_control_view(wc)
        self.assertTrue(control.checkin_allowed())
        self.assertTrue(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

        # On the original you cannot do anything now, because there is a working copy.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

    def test_container_control_normal_folder(self):
        # Folder is not versionable, and not lockable.
        doc = self.portal.docs
        self.assertNotIn("Folder", self.repo.getVersionableContentTypes())
        self.assertFalse(self.repo.isVersionable(doc))
        try:
            lockable = ILockable(doc)
        except TypeError:
            lockable = False
        self.assertFalse(lockable)

        # Nothing is allowed.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

    def test_container_control_folderish_document(self):
        # FolderishDocument is both versionable and lockable.
        doc = self.portal.docs.doc1
        self.assertIn("FolderishDocument", self.repo.getVersionableContentTypes())
        self.assertTrue(self.repo.isVersionable(doc))
        try:
            lockable = ILockable(doc)
        except TypeError:
            lockable = False
        self.assertTrue(lockable)

        # Initially of course you can only check-out.
        control = self.get_control_view(doc)
        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertTrue(control.checkout_allowed())

        # Make a checkout.  Now you can only check-in or cancel.
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        control = self.get_control_view(wc)
        self.assertTrue(control.checkin_allowed())
        self.assertTrue(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

        # On the original you cannot do anything now, because there is a working copy.
        control = self.get_control_view(doc)
        IAnnotations(self.layer["request"]).pop("plone.memoize", None)

        self.assertFalse(control.checkin_allowed())
        self.assertFalse(control.cancel_allowed())
        self.assertFalse(control.checkout_allowed())

    def test_container_control_cancel_on_original_does_not_delete_original(self):
        # checkout document
        doc = self.portal.docs.doc1
        policy = ICheckinCheckoutPolicy(self.portal.docs.doc1, None)
        policy.checkout(self.portal.workarea)

        # get cancel browser view
        from plone.app.iterate.browser.cancel import Cancel

        cancel = Cancel(doc, self.layer["request"])
        self.layer["request"].form["form.button.Cancel"] = True

        # check if cancel on original raises the correct exception
        from plone.app.iterate.interfaces import CheckoutException

        with self.assertRaises(CheckoutException):
            cancel()

    def test_container_relationship_deleted_on_checkin(self):
        # Ensure that the relationship between the baseline and the wc is
        # removed when the working copy is checked in
        folder = self.portal.docs
        doc = folder.doc1

        intids = component.getUtility(IIntIds)
        catalog = component.getUtility(ICatalog)
        obj_id = intids.getId(doc)
        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 0)

        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        wc_id = intids.getId(wc)

        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 1)
        from_rel = rels[0]

        rels = list(catalog.findRelations({"to_id": wc_id}))
        self.assertEqual(len(rels), 1)
        to_rel = rels[0]

        self.assertEqual(from_rel, to_rel)

        doc = ICheckinCheckoutPolicy(wc).checkin("updated")

        rels = list(catalog.findRelations({"from_id": obj_id}))

        self.assertEqual(len(rels), 0)

    def test_relationship_deleted_on_cancel_checkout(self):
        # Ensure that the relationship between the baseline and the wc is
        # removed when the working copy is checked in
        folder = self.portal.docs
        doc = folder.doc1

        intids = component.getUtility(IIntIds)
        catalog = component.getUtility(ICatalog)
        obj_id = intids.getId(doc)
        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 0)

        wc = ICheckinCheckoutPolicy(doc).checkout(folder)
        wc_id = intids.getId(wc)

        rels = list(catalog.findRelations({"from_id": obj_id}))
        self.assertEqual(len(rels), 1)
        from_rel = rels[0]

        rels = list(catalog.findRelations({"to_id": wc_id}))
        self.assertEqual(len(rels), 1)
        to_rel = rels[0]

        self.assertEqual(from_rel, to_rel)

        ICheckinCheckoutPolicy(wc).cancelCheckout()

        rels = list(catalog.findRelations({"from_id": obj_id}))

        self.assertEqual(len(rels), 0)

    def test_relationfield_handling(self):
        # relations are not simply copied
        folder = self.portal.docs
        doc = folder.doc1
        target = folder.doc2
        intids = component.getUtility(IIntIds)
        doc.relatedItems = [RelationValue(intids.getId(target))]
        wc = ICheckinCheckoutPolicy(doc).checkout(doc)
        self.assertNotEqual(doc.relatedItems[0], wc.relatedItems[0])
        self.assertEqual(doc.relatedItems[0].to_object, target)
        self.assertEqual(doc.relatedItems[0].from_object, doc)
        self.assertEqual(wc.relatedItems[0].to_object, target)
        self.assertEqual(wc.relatedItems[0].from_object, wc)
        wc = ICheckinCheckoutPolicy(wc).checkin("modified")
        self.assertEqual(doc.relatedItems[0].to_object, target)
        self.assertEqual(doc.relatedItems[0].from_object, doc)
        self.assertEqual(len(doc.relatedItems), 1)

        wc = ICheckinCheckoutPolicy(doc).checkout(doc)
        wc.relatedItems = [RelationValue(intids.getId(folder))]
        wc = ICheckinCheckoutPolicy(wc).checkin("modified")
        self.assertEqual(doc.relatedItems[0].to_object, folder)
        self.assertEqual(doc.relatedItems[0].from_object, doc)
        self.assertEqual(len(doc.relatedItems), 1)

    def test_publication_behavior_values_not_changed_folderish(self):
        doc = self.portal.docs.doc2
        original_effective_date = doc.effective_date
        original_expiration_date = doc.expiration_date
        # Create a working copy
        wc = ICheckinCheckoutPolicy(doc).checkout(self.portal.workarea)
        # Check in without modifying the existing values
        baseline = ICheckinCheckoutPolicy(wc).checkin("updated")
        # Values should be the same of the original document
        self.assertEqual(original_effective_date, baseline.effective_date)
        self.assertEqual(original_expiration_date, baseline.expiration_date)



"""Testing setup for integration and functional tests."""

from plone.app.contenttypes.testing import PloneAppContenttypes
from plone.app.testing import applyProfile
from plone.app.testing.layers import FunctionalTesting
from plone.app.testing.layers import IntegrationTesting


ADMIN = {
    "id": "admin",
    "password": "secret",
    "roles": ["Manager"],
}
EDITOR = {
    "id": "editor",
    "password": "secret",
    "roles": ["Editor"],
}
CONTRIBUTOR = {
    "id": "contributor",
    "password": "secret",
    "roles": ["Contributor"],
}
USERS_TO_BE_ADDED = (
    ADMIN,
    EDITOR,
    CONTRIBUTOR,
)
USERS_WITH_MEMBER_FOLDER = (
    EDITOR,
    CONTRIBUTOR,
)


class DexPloneAppIterateLayer(PloneAppContenttypes):
    """Dexterity based Plone Sandbox Layer for plone.app.iterate."""

    def setUpZope(self, app, configurationContext):
        """Setup Zope with Addons."""
        super().setUpZope(app, configurationContext)

        import plone.app.iterate
        import plone.app.iterate.tests

        self.loadZCML(package=plone.app.iterate)
        self.loadZCML(package=plone.app.iterate.tests)

    def setUpPloneSite(self, portal):
        """Setup Plone Site with Addons."""
        super().setUpPloneSite(portal)
        applyProfile(portal, "plone.app.iterate:default")
        applyProfile(portal, "plone.app.iterate.tests:testingdx")
        # with named AND dotted behaviors we need to take care of both
        versioning_behavior = {
            "plone.app.versioningbehavior.behaviors.IVersionable",
            "plone.versioning",
        }

        # Disable automatic versioning of Documents.  Note that Documents are
        # still versionable after this.  You just need to do it explicitly,
        # like plone.app.iterate does when needed.
        fti = portal.portal_types["Document"]
        # write back the behaviors without the versioning behaviors
        # using a Set to keep it simple
        # a = set((1,2,3))
        # b = set([2,4])
        # res = tuple(a.difference(b)) >> (1,3)
        fti.behaviors = tuple(
            set(fti.behaviors).difference(versioning_behavior),
        )


PLONEAPPITERATEDEX_FIXTURE = DexPloneAppIterateLayer()
PLONEAPPITERATEDEX_INTEGRATION_TESTING = IntegrationTesting(
    bases=(PLONEAPPITERATEDEX_FIXTURE,), name="DexPloneAppIterateLayer:Integration"
)

PLONEAPPITERATEDEX_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(PLONEAPPITERATEDEX_FIXTURE,), name="DexPloneAppIterateLayer:Functional"
)
